@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %% 
-%% Copyright Ericsson AB 1997-2016. All Rights Reserved.
+%% Copyright Ericsson AB 1997-2017. All Rights Reserved.
 %% 
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -34,21 +34,23 @@
 
 -include("httpd.hrl").
 -include("http_internal.hrl").
--include("httpd_internal.hrl").
+-include_lib("kernel/include/logger.hrl").
 
 -define(HANDSHAKE_TIMEOUT, 5000).
+
 -record(state, {mod,     %% #mod{}
 		manager, %% pid()
 		status,  %% accept | busy | blocked
 		mfa,     %% {Module, Function, Args} 
 		max_keep_alive_request = infinity, %% integer() | infinity
-		response_sent = false, %% true | false 
-		timeout,  %% infinity | integer() > 0
-		timer,     %% ref() - Request timer
-		headers,  %% #http_request_h{}
+		response_sent = false :: boolean(),
+		timeout,   %% infinity | integer() > 0
+		timer      :: 'undefined' | reference(), % Request timer
+		headers,   %% #http_request_h{}
 		body,      %% binary()
 		data,      %% The total data received in bits, checked after 10s
-		byte_limit  %% Bit limit per second before kick out
+		byte_limit, %% Bit limit per second before kick out
+                chunk 
 	       }).
 
 %%====================================================================
@@ -98,21 +100,39 @@ init([Manager, ConfigDB, AcceptTimeout]) ->
     
     {SocketType, Socket} = await_socket_ownership_transfer(AcceptTimeout),
     
+    Peername = http_transport:peername(SocketType, Socket),
+    Sockname = http_transport:sockname(SocketType, Socket),
+ 
     %%Timeout value is in seconds we want it in milliseconds
     KeepAliveTimeOut = 1000 * httpd_util:lookup(ConfigDB, keep_alive_timeout, 150),
     
     case http_transport:negotiate(SocketType, Socket, ?HANDSHAKE_TIMEOUT) of
-	{error, Error} ->
-	    exit({shutdown, Error}); %% Can be 'normal'.
-	ok ->
-	    continue_init(Manager, ConfigDB, SocketType, Socket, KeepAliveTimeOut)
+	{error, {tls_alert, {_, AlertDesc}} = Error} ->
+            ModData = #mod{config_db = ConfigDB, init_data = #init_data{peername = Peername,
+                                                                        sockname = Sockname}},
+            httpd_util:error_log(ConfigDB, httpd_logger:error_report('TLS', AlertDesc, 
+                                                                     ModData, ?LOCATION)),
+	    exit({shutdown, Error}); 
+        {error, _Reason} = Error ->
+            %% This happens if the peer closes the connection 
+            %% or the handshake is timed out. This is not
+            %% an error condition of the server and client will
+            %% retry in the timeout situation.  
+            exit({shutdown, Error}); 
+        {ok, TLSSocket} ->
+	    continue_init(Manager, ConfigDB, SocketType, TLSSocket, 
+                          Peername, Sockname, KeepAliveTimeOut);
+        ok  ->
+            continue_init(Manager, ConfigDB, SocketType, Socket, 
+                          Peername, Sockname, KeepAliveTimeOut)
     end.
 
-continue_init(Manager, ConfigDB, SocketType, Socket, TimeOut) ->
+continue_init(Manager, ConfigDB, SocketType, Socket, Peername, Sockname,
+              TimeOut) ->
     Resolve = http_transport:resolve(),
-    
-    Peername = httpd_socket:peername(SocketType, Socket),
-    InitData = #init_data{peername = Peername, resolve = Resolve},
+    InitData = #init_data{peername = Peername, 
+                          sockname = Sockname, 
+                          resolve = Resolve},
     Mod = #mod{config_db = ConfigDB,
                socket_type = SocketType,
                socket = Socket,
@@ -123,7 +143,8 @@ continue_init(Manager, ConfigDB, SocketType, Socket, TimeOut) ->
     NrOfRequest   = max_keep_alive_request(ConfigDB), 
     MaxContentLen = max_content_length(ConfigDB),
     Customize = customize(ConfigDB),
-
+    MaxChunk = max_client_body_chunk(ConfigDB),
+    
     {_, Status} = httpd_manager:new_connection(Manager),
     
     MFA = {httpd_request, parse, [[{max_uri, MaxURISize}, {max_header, MaxHeaderSize},
@@ -138,7 +159,8 @@ continue_init(Manager, ConfigDB, SocketType, Socket, TimeOut) ->
 		   status                 = Status,
 		   timeout                = TimeOut, 
 		   max_keep_alive_request = NrOfRequest,
-		   mfa                    = MFA},
+		   mfa                    = MFA,
+                   chunk                   = chunk_start(MaxChunk)},
     
     http_transport:setopts(SocketType, Socket, 
 			   [binary, {packet, 0}, {active, once}]),
@@ -159,14 +181,11 @@ continue_init(Manager, ConfigDB, SocketType, Socket, TimeOut) ->
 %%                                      {stop, Reason, State}
 %% Description: Handling call messages
 %%--------------------------------------------------------------------
-handle_call(Request, From, #state{mod = ModData} = State) ->
-    Error = 
-	lists:flatten(
-	  io_lib:format("Unexpected request: "
-			"~n~p"
-			"~nto request handler (~p) from ~p"
-			"~n", [Request, self(), From])),
-    error_log(Error, ModData),
+handle_call(Request, From, #state{mod = #mod{config_db = Db} = ModData} = State) ->
+    httpd_util:error_log(Db, 
+                         httpd_logger:error_report(internal,
+                                                   [{unexpected_call, Request}, {to, self()}, {from, From}], ModData,
+                                                   ?LOCATION)),
     {stop, {call_api_violation, Request, From}, State}.
 
 %%--------------------------------------------------------------------
@@ -175,14 +194,10 @@ handle_call(Request, From, #state{mod = ModData} = State) ->
 %%                                      {stop, Reason, State}
 %% Description: Handling cast messages
 %%--------------------------------------------------------------------
-handle_cast(Msg, #state{mod = ModData} = State) ->
-    Error = 
-	lists:flatten(
-	  io_lib:format("Unexpected message: "
-			"~n~p"
-			"~nto request handler (~p)"
-			"~n", [Msg, self()])),
-    error_log(Error, ModData),
+handle_cast(Msg, #state{mod = #mod{config_db = Db} = ModData} = State) ->
+    httpd_util:error_log(Db,
+                         httpd_logger:error_report(internal, [{unexpected_cast, Msg}, {to, self()}], ModData,
+                                                   ?LOCATION)),
     {noreply, State}.
 
 %%--------------------------------------------------------------------
@@ -193,6 +208,7 @@ handle_cast(Msg, #state{mod = ModData} = State) ->
 %%--------------------------------------------------------------------
 handle_info({Proto, Socket, Data}, 
 	    #state{mfa = {Module, Function, Args},
+                   chunk = {ChunkState, _},
 		   mod = #mod{socket_type = SockType, 
 			      socket = Socket} = ModData} = State) 
   when (((Proto =:= tcp) orelse 
@@ -206,7 +222,8 @@ handle_info({Proto, Socket, Data},
 		      _ ->
 			  State#state.data + byte_size(Data)
 		  end,
-    case PROCESSED of
+
+    case PROCESSED of       
         {ok, Result} ->
 	    NewState = case NewDataSize of
 			   undefined ->
@@ -214,16 +231,16 @@ handle_info({Proto, Socket, Data},
 			   _ ->
 			       set_new_data_size(cancel_request_timeout(State), NewDataSize)
 		       end,
-            handle_http_msg(Result, NewState); 
+            handle_msg(Result, NewState);
 	{error, {size_error, MaxSize, ErrCode, ErrStr}, Version} ->
 	    NewModData =  ModData#mod{http_version = Version},
-	    httpd_response:send_status(NewModData, ErrCode, ErrStr),
-	    Reason = io_lib:format("~p: ~p max size is ~p~n", 
-				   [ErrCode, ErrStr, MaxSize]),
-	    error_log(Reason, NewModData),
+	    httpd_response:send_status(NewModData, ErrCode, ErrStr, {max_size, MaxSize}),
 	    {stop, normal, State#state{response_sent = true, 
 				       mod = NewModData}};
-
+        
+        {http_chunk = Module, Function, Args} when ChunkState =/= undefined ->
+            NewState = handle_chunk(Module, Function, Args, State),
+            {noreply, NewState};
 	NewMFA ->
 	    http_transport:setopts(SockType, Socket, [{active, once}]),
 	    case NewDataSize of
@@ -240,20 +257,18 @@ handle_info({tcp_closed, _}, State) ->
 handle_info({ssl_closed, _}, State) ->
     {stop, normal, State};
 handle_info({tcp_error, _, _} = Reason, State) ->
-    {stop, Reason, State};
+    {stop, {shutdown, Reason}, State};
 handle_info({ssl_error, _, _} = Reason, State) ->
-    {stop, Reason, State};
+    {stop, {shutdown, Reason}, State};
 
 %% Timeouts
 handle_info(timeout, #state{mfa = {_, parse, _}} = State) ->
-    %% error_log("No request received on keep-alive connection "
-    %% 	      "before server side timeout", ModData),
-    %% No response should be sent!
+    %% No request received on keep-alive connection 
+    %% before server side timeout. No response should be sent!
     {stop, normal, State#state{response_sent = true}}; 
 handle_info(timeout, #state{mod = ModData} = State) ->
-    httpd_response:send_status(ModData, 408, "Request timeout"),
-    error_log("The client did not send the whole request before the "
-	      "server side timeout", ModData),
+    httpd_response:send_status(ModData, 408, "Request timeout", "The client did not send the whole request before the "
+                               "server side timeout"),
     {stop, normal, State#state{response_sent = true}};
 handle_info(check_data_first, #state{data = Data, byte_limit = Byte_Limit} = State) ->
     case Data >= (Byte_Limit*3) of 
@@ -276,13 +291,11 @@ handle_info({'EXIT', _, Reason}, State) ->
     {stop, Reason, State};
 
 %% Default case
-handle_info(Info, #state{mod = ModData} = State) ->
-    Error = lists:flatten(
-	      io_lib:format("Unexpected info: "
-			    "~n~p"
-			    "~nto request handler (~p)"
-			    "~n", [Info, self()])),
-    error_log(Error, ModData),
+handle_info(Info, #state{mod = #mod{config_db = Db} =ModData} = State) ->
+    httpd_util:error_log(Db, 
+                         httpd_logger:error_report(internal,
+                                                   [{unexpected_info, Info}, {to, self()}], ModData,
+                                                   ?LOCATION)),
     {noreply, State}.
 
 
@@ -301,10 +314,6 @@ terminate({shutdown,_}, State) ->
     do_terminate(State);
 terminate(Reason, #state{response_sent = false, mod = ModData} = State) ->
     httpd_response:send_status(ModData, 500, none),
-     ReasonStr = 
-	lists:flatten(io_lib:format("~s - ~p", 
-				    [httpd_util:reason_phrase(500), Reason])),
-    error_log(ReasonStr, ModData),
     terminate(Reason, State#state{response_sent = true, mod = ModData});
 terminate(_Reason, State) ->
     do_terminate(State).
@@ -348,6 +357,34 @@ await_socket_ownership_transfer(AcceptTimeout) ->
 	    exit(accept_socket_timeout)
     end.
 
+
+%%% Internal chunking of client body 
+handle_msg({{continue, Chunk}, Module, Function, Args}, #state{chunk = {_, CbState}} = State) ->
+    handle_internal_chunk(State#state{chunk = {continue, CbState},
+                                      body = Chunk}, Module, Function, Args);
+handle_msg({continue, Module, Function, Args}, 	#state{mod = ModData} = State) ->
+    http_transport:setopts(ModData#mod.socket_type, 
+                           ModData#mod.socket, 
+                           [{active, once}]),
+    {noreply, State#state{mfa = {Module, Function, Args}}};
+handle_msg({last, Body}, #state{headers = Headers, chunk = {_, CbState}} = State) -> 
+    NewHeaders = Headers#http_request_h{'content-length' = integer_to_list(size(Body))},
+    handle_response(State#state{chunk = {last, CbState},
+                                headers = NewHeaders,
+                                body = Body});
+%%% Last data chunked by client
+handle_msg({ChunkedHeaders, Body}, #state{headers = Headers , chunk = {ChunkState, CbState}} = State) when ChunkState =/= undefined ->
+    NewHeaders = http_chunk:handle_headers(Headers, ChunkedHeaders),
+    handle_response(State#state{chunk = {last, CbState},
+                                headers = NewHeaders,
+                                body = Body});
+handle_msg({ChunkedHeaders, Body}, #state{headers = Headers , chunk = {undefined, _}} = State) ->
+    NewHeaders = http_chunk:handle_headers(Headers, ChunkedHeaders),
+    handle_response(State#state{headers = NewHeaders,
+                                body = Body});
+handle_msg(Result, State) ->
+    handle_http_msg(Result, State).
+
 handle_http_msg({_, _, Version, {_, _}, _}, 
 		#state{status = busy, mod = ModData} = State) -> 
     handle_manager_busy(State#state{mod = 
@@ -363,9 +400,9 @@ handle_http_msg({_, _, Version, {_, _}, _},
 handle_http_msg({Method, Uri, Version, {RecordHeaders, Headers}, Body},
 		#state{status = accept, mod = ModData} = State) ->        
     case httpd_request:validate(Method, Uri, Version) of
-	ok  ->
+	{ok, NormalizedURI}  ->
 	    {ok, NewModData} = 
-		httpd_request:update_mod_data(ModData, Method, Uri,
+		httpd_request:update_mod_data(ModData, Method, NormalizedURI,
 					      Version, Headers),
       
 	    case is_host_specified_if_required(NewModData#mod.absolute_uri,
@@ -382,32 +419,16 @@ handle_http_msg({Method, Uri, Version, {RecordHeaders, Headers}, Body},
 	    end;
 	{error, {not_supported, What}} ->
 	    httpd_response:send_status(ModData#mod{http_version = Version},
-				       501, {Method, Uri, Version}),
-	    Reason = io_lib:format("Not supported: ~p~n", [What]),
-	    error_log(Reason, ModData),
-	    {stop, normal, State#state{response_sent = true}};
-	{error, {bad_request, {forbidden, URI}}} ->
-	    httpd_response:send_status(ModData#mod{http_version = Version},
-				       403, URI),
-	    Reason = io_lib:format("Forbidden URI: ~p~n", [URI]),
-	    error_log(Reason, ModData),
+				       501, {Method, Uri, Version}, {not_sup, What}),
 	    {stop, normal, State#state{response_sent = true}};
 	{error, {bad_request, {malformed_syntax, URI}}} ->
 	    httpd_response:send_status(ModData#mod{http_version = Version},
-				       400, URI),
-	    Reason = io_lib:format("Malformed syntax in URI: ~p~n", [URI]),
-	    error_log(Reason, ModData),
+				       400, URI, {malformed_syntax, URI}),
 	    {stop, normal, State#state{response_sent = true}};
 	{error, {bad_version, Ver}} ->
-	    httpd_response:send_status(ModData#mod{http_version = "HTTP/0.9"}, 400, Ver),
-	    Reason = io_lib:format("Malformed syntax version: ~p~n", [Ver]),
-	    error_log(Reason, ModData),
+	    httpd_response:send_status(ModData#mod{http_version = "HTTP/0.9"}, 400, Ver, {malformed_syntax, Ver}),
 	    {stop, normal, State#state{response_sent = true}}
     end;
-handle_http_msg({ChunkedHeaders, Body}, 
-		State = #state{headers = Headers}) ->
-    NewHeaders = http_chunk:handle_headers(Headers, ChunkedHeaders),
-    handle_response(State#state{headers = NewHeaders, body = Body});
 handle_http_msg(Body, State) ->
     handle_response(State#state{body = Body}).
 
@@ -442,58 +463,70 @@ handle_body(#state{mod = #mod{config_db = ConfigDB}} = State) ->
     
     end.
 	
-handle_body(#state{headers = Headers, body = Body, mod = ModData} = State,
+handle_body(#state{headers = Headers, body = Body, 
+                   chunk =  {ChunkState, CbState}, mod = #mod{config_db = ConfigDB} = ModData} = State,
 	    MaxHeaderSize, MaxBodySize) ->
+    MaxChunk = max_client_body_chunk(ConfigDB),
     case Headers#http_request_h.'transfer-encoding' of
 	"chunked" ->
 	    try http_chunk:decode(Body, MaxBodySize, MaxHeaderSize) of
-		{Module, Function, Args} ->
+                {Module, Function, Args} ->
 		    http_transport:setopts(ModData#mod.socket_type, 
 					   ModData#mod.socket, 
 					   [{active, once}]),
 		    {noreply, State#state{mfa = 
-					  {Module, Function, Args}}};
-		{ok, {ChunkedHeaders, NewBody}} ->
-		    NewHeaders = 
-			http_chunk:handle_headers(Headers, ChunkedHeaders),
-		    handle_response(State#state{headers = NewHeaders,
-						body = NewBody})
+                                              {Module, Function, Args},
+                                          chunk = chunk_start(MaxChunk)}};
+                {ok, {ChunkedHeaders, NewBody}} ->
+		    NewHeaders = http_chunk:handle_headers(Headers, ChunkedHeaders),	
+                    handle_response(State#state{headers = NewHeaders,
+                                                body = NewBody,
+                                                chunk = chunk_finish(ChunkState, CbState, MaxChunk)})
 	    catch 
 		throw:Error ->
 		    httpd_response:send_status(ModData, 400, 
-					       "Bad input"),
-		    Reason = io_lib:format("Chunk decoding failed: ~p~n", 
-					   [Error]),
-		    error_log(Reason, ModData),
+					       "Bad input", {chunk_decoding, bad_input, Error}),
 		    {stop, normal, State#state{response_sent = true}}  
 	    end;
 	Encoding when is_list(Encoding) ->
 	    httpd_response:send_status(ModData, 501, 
-				       "Unknown Transfer-Encoding"),
-	    Reason = io_lib:format("Unknown Transfer-Encoding: ~p~n", 
-				   [Encoding]),
-	    error_log(Reason, ModData),
+				       "Unknown Transfer-Encoding", 
+                                       {unknown_transfer_encoding, Encoding}),
 	    {stop, normal, State#state{response_sent = true}};
 	_ -> 
-	    Length = list_to_integer(Headers#http_request_h.'content-length'),	    
+	    Length = list_to_integer(Headers#http_request_h.'content-length'),
+	    MaxChunk = max_client_body_chunk(ConfigDB),
 	    case ((Length =< MaxBodySize) or (MaxBodySize == nolimit)) of
 		true ->
-		    case httpd_request:whole_body(Body, Length) of 
-			{Module, Function, Args} ->
-			    http_transport:setopts(ModData#mod.socket_type, 
+		    case httpd_request:body_chunk_first(Body, Length, MaxChunk) of 
+                        %% This is the case that the we need more data to complete
+                        %% the body but chunking to the mod_esi user is not enabled.
+                        {Module, add_chunk = Function,  Args} ->  
+                            http_transport:setopts(ModData#mod.socket_type, 
 						   ModData#mod.socket, 
 						   [{active, once}]),
 			    {noreply, State#state{mfa = 
 						      {Module, Function, Args}}};
-			
-			{ok, NewBody} ->
-			    handle_response(
-			      State#state{headers = Headers,
-					  body = NewBody})
+                        %% Chunking to mod_esi user is enabled
+                        {ok, {continue, Module, Function, Args}} ->
+                                http_transport:setopts(ModData#mod.socket_type, 
+						   ModData#mod.socket, 
+						   [{active, once}]),
+			    {noreply, State#state{mfa = 
+						      {Module, Function, Args}}};
+                        {ok, {{continue, Chunk}, Module, Function, Args}} ->
+                            handle_internal_chunk(State#state{chunk =  chunk_start(MaxChunk), 
+                                                              body = Chunk}, Module, Function, Args);                   
+                        %% Whole body delivered, if chunking mechanism is enabled the whole
+                        %% body fits in one chunk.
+                        {ok, NewBody} ->
+                            handle_response(State#state{chunk = chunk_finish(ChunkState, 
+                                                                             CbState, MaxChunk),
+                                                        headers = Headers,
+                                                        body = NewBody})
 		    end;
 		false ->
 		    httpd_response:send_status(ModData, 413, "Body too long"),
-		    error_log("Body too long", ModData),
 		    {stop, normal,  State#state{response_sent = true}}
 	    end
     end.
@@ -508,22 +541,21 @@ handle_expect(#state{headers = Headers, mod =
 	    ok;
 	continue when MaxBodySize < Length ->
 	    httpd_response:send_status(ModData, 413, "Body too long"),
-	    error_log("Body too long", ModData),
 	    {stop, normal, State#state{response_sent = true}};
 	{break, Value} ->
 	    httpd_response:send_status(ModData, 417, 
-				       "Unexpected expect value"),
-	    Reason = io_lib:format("Unexpected expect value: ~p~n", [Value]),
-	    error_log(Reason, ModData),
+				       "Unexpected expect value",
+                                       {unexpected, Value}
+                                      ),
 	    {stop, normal,  State#state{response_sent = true}};
 	no_expect_header ->
 	    ok;
 	http_1_0_expect_header ->
 	    httpd_response:send_status(ModData, 400, 
 				       "Only HTTP/1.1 Clients "
-				       "may use the Expect Header"),
-	    error_log("Client with lower version than 1.1 tried to send"
-		      "an expect header", ModData),
+				       "may use the Expect Header", 
+                                       "Client with lower version than 1.1 tried to send"
+                                       "an expect header"),
 	    {stop, normal, State#state{response_sent = true}}
     end.
 
@@ -549,15 +581,61 @@ expect(Headers, _, ConfigDB) ->
 	    end
     end.
 
+handle_chunk(http_chunk = Module, decode_data = Function, 
+             [ChunkSize, TotalChunk, {MaxBodySize, BodySoFar, _AccLength, MaxHeaderSize}],
+             #state{chunk = {_, CbState},
+                    mod = #mod{socket_type = SockType,
+                               socket = Socket} = ModData} = State) ->
+    {continue, NewCbState} = httpd_response:handle_continuation(ModData#mod{entity_body = 
+                                                                                {continue, BodySoFar, CbState}}),
+    http_transport:setopts(SockType, Socket, [{active, once}]),
+    State#state{chunk = {continue, NewCbState}, mfa = {Module, Function, [ChunkSize, TotalChunk, {MaxBodySize, <<>>, 0, MaxHeaderSize}]}};
+
+handle_chunk(http_chunk = Module, decode_size = Function, 
+             [Data, HexList, _AccSize, {MaxBodySize, BodySoFar, _AccLength, MaxHeaderSize}],
+             #state{chunk = {_, CbState},
+                    mod = #mod{socket_type = SockType,
+                               socket = Socket} = ModData} = State) ->
+    {continue, NewCbState} = httpd_response:handle_continuation(ModData#mod{entity_body = {continue, BodySoFar, CbState}}),
+    http_transport:setopts(SockType, Socket, [{active, once}]),
+    State#state{chunk = {continue, NewCbState}, mfa = {Module, Function, [Data, HexList, 0, {MaxBodySize, <<>>, 0, MaxHeaderSize}]}};
+handle_chunk(Module, Function, Args, #state{mod = #mod{socket_type = SockType,
+                                                                      socket = Socket}} = State) ->
+    http_transport:setopts(SockType, Socket, [{active, once}]),
+    State#state{mfa = {Module, Function, Args}}.
+
+handle_internal_chunk(#state{chunk = {ChunkState, CbState}, body = Chunk, 
+                             mod = #mod{socket_type = SockType,
+                                        socket = Socket} = ModData} = State, Module, Function, Args)->
+    Bodychunk = body_chunk(ChunkState, CbState, Chunk),
+    {continue, NewCbState} = httpd_response:handle_continuation(ModData#mod{entity_body = Bodychunk}),
+    case Args of
+        [<<>> | _] ->
+            http_transport:setopts(SockType, Socket, [{active, once}]),
+            {noreply, State#state{chunk = {continue, NewCbState}, mfa = {Module, Function, Args}}};
+        _ ->
+            handle_info({dummy, Socket, <<>>}, State#state{chunk = {continue, NewCbState}, 
+                                                           mfa = {Module, Function, Args}})
+    end.
+
+handle_response(#state{body    = Body, 
+                       headers = Headers,
+		       mod     = ModData, 
+                       chunk   = {last, CbState},
+		       max_keep_alive_request = Max} = State) when Max > 0 ->
+    {NewBody, Data} = httpd_request:body_data(Headers, Body),
+    ok = httpd_response:generate_and_send_response(
+           ModData#mod{entity_body = {last, NewBody, CbState}}),     
+    handle_next_request(State#state{response_sent = true}, Data);
 handle_response(#state{body    = Body, 
 		       mod     = ModData, 
 		       headers = Headers,
 		       max_keep_alive_request = Max} = State) when Max > 0 ->
     {NewBody, Data} = httpd_request:body_data(Headers, Body),
+    %% Backwards compatible, may cause memory explosion
     ok = httpd_response:generate_and_send_response(
-	   ModData#mod{entity_body = NewBody}),
+           ModData#mod{entity_body = binary_to_list(NewBody)}),
     handle_next_request(State#state{response_sent = true}, Data);
-
 handle_response(#state{body    = Body, 
 		       headers = Headers, 
 		       mod     = ModData} = State) ->
@@ -577,6 +655,7 @@ handle_next_request(#state{mod = #mod{connection = true} = ModData,
     MaxURISize    = max_uri_size(ModData#mod.config_db), 
     MaxContentLen = max_content_length(ModData#mod.config_db),
     Customize = customize(ModData#mod.config_db),
+    MaxChunk = max_client_body_chunk(ModData#mod.config_db),
 
     MFA = {httpd_request, parse, [[{max_uri, MaxURISize}, {max_header, MaxHeaderSize},
 				   {max_version, ?HTTP_MAX_VERSION_STRING}, 
@@ -589,6 +668,7 @@ handle_next_request(#state{mod = #mod{connection = true} = ModData,
 			   max_keep_alive_request = decrease(Max),
 			   headers                = undefined, 
 			   body                   = undefined,
+                           chunk                  = chunk_start(MaxChunk),
 			   response_sent          = false},
     
     NewState = activate_request_timeout(TmpState),
@@ -633,19 +713,16 @@ decrease(N) when is_integer(N) ->
 decrease(N) ->
     N.
 
-error_log(ReasonString,  #mod{config_db = ConfigDB}) ->
-    Error = lists:flatten(
-	      io_lib:format("Error reading request: ~s", [ReasonString])),
-    httpd_util:error_log(ConfigDB, Error).
-
-
-%%--------------------------------------------------------------------
+%--------------------------------------------------------------------
 %% Config access wrapper functions
 %%--------------------------------------------------------------------
 
 max_header_size(ConfigDB) ->
     httpd_util:lookup(ConfigDB, max_header_size, ?HTTP_MAX_HEADER_SIZE).
 
+max_client_body_chunk(ConfigDB) ->
+    httpd_util:lookup(ConfigDB, max_client_body_chunk, nolimit).
+
 max_uri_size(ConfigDB) ->
     httpd_util:lookup(ConfigDB, max_uri_size, ?HTTP_MAX_URI_SIZE).
 
@@ -660,3 +737,17 @@ max_content_length(ConfigDB) ->
 
 customize(ConfigDB) ->    
     httpd_util:lookup(ConfigDB, customize, httpd_custom).
+
+chunk_start(nolimit) ->
+    {undefined, undefined};
+chunk_start(_) ->
+    {first, undefined}.
+chunk_finish(_, _, nolimit) ->
+    {undefined, undefined};
+chunk_finish(_, CbState, _) ->
+    {last, CbState}.
+
+body_chunk(first, _, Chunk) ->
+    {first, Chunk};
+body_chunk(ChunkState, CbState, Chunk) ->
+    {ChunkState, Chunk, CbState}.
