@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1999-2018. All Rights Reserved.
+%% Copyright Ericsson AB 1999-2020. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -116,6 +116,10 @@ dflag2str(?DFLAG_SEND_SENDER) ->
     "SEND_SENDER";
 dflag2str(?DFLAG_BIG_SEQTRACE_LABELS) ->
     "BIG_SEQTRACE_LABELS";
+dflag2str(?DFLAG_EXIT_PAYLOAD) ->
+    "EXIT_PAYLOAD";
+dflag2str(?DFLAG_FRAGMENTS) ->
+    "FRAGMENTS";
 dflag2str(_) ->
     "UNKNOWN".
 
@@ -340,22 +344,11 @@ shutdown(_Module, _Line, _Data, Reason) ->
     ?shutdown_trace("Net Kernel 2: shutting down connection "
 		    "~p:~p, data ~p,reason ~p~n",
 		    [_Module,_Line, _Data, Reason]),
-    flush_down(),
     exit(Reason).
 %% Use this line to debug connection.  
 %% Set net_kernel verbose = 1 as well.
 %%    exit({Reason, ?MODULE, _Line, _Data, erlang:timestamp()}).
 
-
-flush_down() ->
-    receive
-	{From, get_status} ->
-	    From ! {self(), get_status, error},
-	    flush_down()
-    after 0 ->
-	    ok
-    end.
-
 handshake_we_started(#hs_data{request_type=ReqType,
 			      other_node=Node,
                               add_flags=AddFlgs0,
